@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import functools
 import itertools
 import random
@@ -19,7 +21,20 @@ from adaptive.learner.triangulation import (
     simplex_volume_in_embedding,
 )
 from adaptive.notebook_integration import ensure_holoviews, ensure_plotly
-from adaptive.utils import cache_latest, restore
+from adaptive.utils import (
+    assign_defaults,
+    cache_latest,
+    partial_function_from_dataframe,
+    restore,
+)
+
+try:
+    import pandas
+
+    with_pandas = True
+
+except ModuleNotFoundError:
+    with_pandas = False
 
 
 def to_list(inp):
@@ -361,6 +376,10 @@ class LearnerND(BaseLearner):
         # _pop_highest_existing_simplex
         self._simplex_queue = SortedKeyList(key=_simplex_evaluation_priority)
 
+    def new(self) -> LearnerND:
+        """Create a new learner with the same function and bounds."""
+        return LearnerND(self.function, self.bounds, self.loss_per_simplex)
+
     @property
     def npoints(self):
         """Number of evaluated points."""
@@ -388,6 +407,87 @@ class LearnerND(BaseLearner):
         of ``learner.function``."""
         return np.array([(*p, *np.atleast_1d(v)) for p, v in sorted(self.data.items())])
 
+    def to_dataframe(
+        self,
+        with_default_function_args: bool = True,
+        function_prefix: str = "function.",
+        point_names: tuple[str, ...] = ("x", "y", "z"),
+        value_name: str = "value",
+    ) -> pandas.DataFrame:
+        """Return the data as a `pandas.DataFrame`.
+
+        Parameters
+        ----------
+        with_default_function_args : bool, optional
+            Include the ``learner.function``'s default arguments as a
+            column, by default True
+        function_prefix : str, optional
+            Prefix to the ``learner.function``'s default arguments' names,
+            by default "function."
+        point_names : tuple[str, ...], optional
+            Names of the input points, should be the same length as number
+            of input parameters by default ("x", "y", "z" )
+        value_name : str, optional
+            Name of the output value, by default "value"
+
+        Returns
+        -------
+        pandas.DataFrame
+
+        Raises
+        ------
+        ImportError
+            If `pandas` is not installed.
+        """
+        if not with_pandas:
+            raise ImportError("pandas is not installed.")
+        if len(point_names) != self.ndim:
+            raise ValueError(
+                f"point_names ({point_names}) should have the"
+                f" same length as learner.ndims ({self.ndim})"
+            )
+        data = list((*x, y) for x, y in self.data.items())
+        df = pandas.DataFrame(data, columns=[*point_names, value_name])
+        df.attrs["inputs"] = list(point_names)
+        df.attrs["output"] = value_name
+        if with_default_function_args:
+            assign_defaults(self.function, df, function_prefix)
+        return df
+
+    def load_dataframe(
+        self,
+        df: pandas.DataFrame,
+        with_default_function_args: bool = True,
+        function_prefix: str = "function.",
+        point_names: tuple[str, ...] = ("x", "y", "z"),
+        value_name: str = "value",
+    ):
+        """Load data from a `pandas.DataFrame`.
+
+        If ``with_default_function_args`` is True, then ``learner.function``'s
+        default arguments are set (using `functools.partial`) from the values
+        in the `pandas.DataFrame`.
+
+        Parameters
+        ----------
+        df : pandas.DataFrame
+            The data to load.
+        with_default_function_args : bool, optional
+            The ``with_default_function_args`` used in ``to_dataframe()``,
+            by default True
+        function_prefix : str, optional
+            The ``function_prefix`` used in ``to_dataframe``, by default "function."
+        point_names : str, optional
+            The ``point_names`` used in ``to_dataframe``, by default ("x", "y", "z")
+        value_name : str, optional
+            The ``value_name`` used in ``to_dataframe``, by default "value"
+        """
+        self.tell_many(df[list(point_names)].values, df[value_name].values)
+        if with_default_function_args:
+            self.function = partial_function_from_dataframe(
+                self.function, df, function_prefix
+            )
+
     @property
     def bounds_are_done(self):
         return all(p in self.data for p in self._bounds_points)
